/*===============================================================*/

#include <stdio.h>
#include <assert.h>
#include "parti.h"

/*==================================================================*/

HASHSTR * PARTI_create_hash_table( data_range )

int        data_range;
{
    int      i,j;
    int      hsize, estimate_hash_table_size() ;
    HASHSTR *table;

    table =  ( HASHSTR *)malloc(sizeof(struct hashtable) );
    assert( table != NULL );
    hsize             =  estimate_hash_table_size( data_range ) ;
    assert( hsize != 0 );

    table->type		= DEFHASH;	/* default type of hash table */
    table->first_stamp  = 0;
    table->first_offset = 0;
    table->no_buckets   = hsize ;
    table->no_stamps    = 0 ;
    table->stamp_mask   = 0 ;

    for ( i = 0; i< MAX_STAMPS; i++ ) {
       table->size_stamp[i]   = 0;
       table->area_stamp[i]   = 0;
       table->start_offset[i] = 0;
    }

    table->buckets    = (HASHCELL *)malloc(sizeof(struct hashcell)*hsize);
    assert( table->buckets != NULL );

    for(i=0; i<hsize; i++) {
        table->buckets[i].next = NULL ;
        table->buckets[i].tag  = 0;
	for( j = 0; j< CHAINSIZE; j++ ) {
	  table->buckets[i].key[j]   = -1;
	  table->buckets[i].stamp[j] =  0;
        }
    }

    return (table) ;
}

/*==================================================================*/

HASHSTR * PARTI_create_rep_hash_table( trans_table )
TTABLE	*trans_table;

{
    int      i,j, ndata;
    HASHSTR *table;

    table =  ( HASHSTR *)malloc(sizeof(struct hashtable) );
    assert( table != NULL );

    ndata		= trans_table->tabptr->pageSize;
    table->type		= REPHASH;	/* hash table for replicated ttable */
    table->first_stamp  = 0;
    table->first_offset = 0;
    table->no_buckets   = ndata;
    table->no_stamps    = 0 ;
    table->stamp_mask   = 0 ;
    table->trans_table	= trans_table;

    for ( i = 0; i< MAX_STAMPS; i++ ) {
       table->size_stamp[i]   = 0;
       table->area_stamp[i]   = 0;
       table->start_offset[i] = 0;
    }

    table->cells   = (REPHCELL *)malloc(sizeof(REPHCELL)*ndata);
    assert( table->cells != NULL );

    for(i=0; i<ndata; i++) {
        table->cells[i].offset = NULL ;
        table->cells[i].stamp  = 0;
    }

    return (table) ;
}

/*=======================================================================*/
void  PARTI_free_hash_table( hash_table )

HASHSTR *hash_table;
{
    int       no_buckets, i;
    HASHCELL  *hashcell, *tmp;

    if (hash_table->type == REPHASH) {
	free (hash_table->cells);
	free (hash_table);
	return;
    }

    no_buckets = hash_table->no_buckets;
    for ( i=0; i<no_buckets; i++ ) {
        hashcell = &(hash_table->buckets[i]);
        while( hashcell != NULL ) {
            tmp = hashcell->next ;
            free(hashcell);
            hashcell = tmp ;
        }
    }
    free( hash_table );
    return;
}


/*=======================================================================*/
void    PARTI_hash( trans_table, hash_table, supermask, new_stamp,
                    global_refs,  ndata, my_size,
                    no_off_proc, no_new_off_proc  )

TTABLE   *trans_table;
HASHSTR  *hash_table;
int       supermask;
int      *new_stamp;
int      *global_refs;
int       ndata;
int       my_size;
int      *no_off_proc ;
int      *no_new_off_proc ;

{
  int      i, stamp;
  int      *proc,  *local ;
  double   dclock(), end, st ;
  int      *tmp1 ,*tmp2, a,b,c;


  stamp = ((*new_stamp) == NO_STAMP) ?  NO_STAMP : 0;

  if (tableGetReplication(trans_table) >= 1.0) {
    enter_rep_hash (trans_table, global_refs, ndata, hash_table, 
						&stamp, supermask, my_size);

  }
  else {
    proc          = (int * ) malloc ( sizeof(int) * ndata );
    local         = (int * ) malloc ( sizeof(int) * ndata ) ;
    assert(local !=  NULL && proc != NULL);

    dereference( trans_table, global_refs, local, proc,  ndata );

    enter_hash( global_refs, ndata, 
              hash_table, &stamp, supermask,  proc,  local, my_size);

    free( proc) ;
    free( local );
  }
  
  (*new_stamp)      = ((*new_stamp) != NO_STAMP)?stamp:NO_STAMP ;
  (*no_off_proc)    = hash_table->size_stamp[stamp] ;
  (*no_new_off_proc)= hash_table->area_stamp[stamp] ;

  return ;
} 

/*=======================================================================*/

void     PARTI_schedule( hash_table, combo_mask, sched, maxdim)

HASHSTR  *hash_table ;
int       combo_mask;
SCHED   **sched;
int       maxdim;
{

   int     *fetch_p, *fetch_l, *glb, *lcl, fetch_size; 
   int      siz;
   int      contiguous = 0 ;
   double   st, end;

   siz        = find_size( hash_table, combo_mask);
   contiguous = check_contiguity( hash_table, combo_mask );

   fetch_p = (int *)malloc( sizeof(int)*siz ); 
   fetch_l = (int *)malloc( sizeof(int)*siz ); 
   glb     = (int *)malloc( sizeof(int)*siz ); 
   lcl     = (int *)malloc( sizeof(int)*siz ); 
   assert( fetch_p!= NULL && fetch_l != NULL && glb != NULL && lcl != NULL);

   if ( contiguous ) {
      if (hash_table->type == DEFHASH)
         PARTI_inc_fetch_lists( hash_table, combo_mask,fetch_p, fetch_l,
    				  glb,lcl, &fetch_size);
      else if (hash_table->type == REPHASH)
	 PARTI_rep_inc_fetch_lists (hash_table, combo_mask,
					fetch_p, fetch_l, glb,lcl, &fetch_size);
      else
	 printf ("Error: hash table\n");
   }
   else {
      if (hash_table->type == DEFHASH)
	 PARTI_fetch_lists( hash_table, combo_mask ,fetch_p, fetch_l, 
        		   glb,lcl, &fetch_size);
      else if (hash_table->type == REPHASH)
 	 PARTI_rep_fetch_lists(hash_table, combo_mask,
					fetch_p, fetch_l, glb,lcl, &fetch_size);
      else
	 printf ("Error: hash table\n");
   }				

   (*sched)     =  schedulen( fetch_l, fetch_p, glb, lcl, fetch_size,  maxdim);
   (*sched)->type =  ( contiguous ? NORMAL : NC);  

   free( fetch_p); free( fetch_l); free( glb); free( lcl );
   return ;
}

/*===================================================================*/
void     PARTI_schedule_nc( hash_table, combo_mask, sched, maxdim)

HASHSTR  *hash_table ;
int       combo_mask;
SCHED   **sched;
int       maxdim;
{

   int     *fetch_p, *fetch_l, *glb, *lcl, fetch_size; 
   int      siz;
   int      contiguous = 0 ;
   double   st, end;

   siz        = find_size( hash_table, combo_mask);

   fetch_p = (int *)malloc( sizeof(int)*siz ); 
   fetch_l = (int *)malloc( sizeof(int)*siz ); 
   glb     = (int *)malloc( sizeof(int)*siz ); 
   lcl     = (int *)malloc( sizeof(int)*siz ); 
   assert( fetch_p!= NULL && fetch_l != NULL && glb != NULL && lcl != NULL);

      if (hash_table->type == DEFHASH)
	 PARTI_fetch_lists( hash_table, combo_mask ,fetch_p, fetch_l, 
        		   glb,lcl, &fetch_size);
      else if (hash_table->type == REPHASH)
 	 PARTI_rep_fetch_lists(hash_table, combo_mask,
					fetch_p, fetch_l, glb,lcl, &fetch_size);
      else
	 printf ("Error: hash table\n");

   (*sched)     =  schedulen( fetch_l, fetch_p, glb, lcl, fetch_size,  maxdim);
   (*sched)->type =  (NC);  

   free( fetch_p); free( fetch_l); free( glb); free( lcl );
   return ;
}
/*=================================================================*/


void  PARTI_incremental_schedule( hash_table, combo_mask, sched, maxdim)

HASHSTR  *hash_table ;
int       combo_mask;
SCHED   **sched;
int       maxdim;
{

   int     *fetch_p, *fetch_l, *glb, *lcl, fetch_size; 
   int      siz;
   int      contiguous = 0 ;
   double   st, end;
   void     find_size_and_contiguity();  

   siz = find_inc_size( hash_table, combo_mask);

   fetch_p = (int *)malloc( sizeof(int)*siz ); 
   fetch_l = (int *)malloc( sizeof(int)*siz ); 
   glb     = (int *)malloc( sizeof(int)*siz ); 
   lcl     = (int *)malloc( sizeof(int)*siz ); 
   assert( fetch_p!= NULL && fetch_l != NULL && glb != NULL && lcl != NULL);

   if (hash_table->type == DEFHASH)
     PARTI_inc_fetch_lists( hash_table, combo_mask,fetch_p, fetch_l,
  				  glb,lcl, &fetch_size);
   else if (hash_table->type == REPHASH)
     PARTI_rep_inc_fetch_lists (hash_table, combo_mask,
				fetch_p, fetch_l, glb,lcl, &fetch_size);
/*
   PARTI_inc_fetch_lists( hash_table, combo_mask ,fetch_p, fetch_l, 
					      glb,lcl, &fetch_size);
*/

   (*sched)   =  schedulen( fetch_l, fetch_p, glb, lcl, fetch_size,  maxdim);

   free( fetch_p); free( fetch_l); free( glb); free( lcl );
   return ; 
}

/*===================================================================*/

int   PARTI_make_mask( stamp )
int   stamp ;
{
    int test=1;
    if (!( stamp < 32 && stamp >= 0)) {
      test = test;
    }
    test = (test << stamp);
    return (test );
}

/*===================================================================*/

void  PARTI_clear_mask ( hash_table, mask )
HASHSTR *hash_table ;
int   mask ;
{
     clear_mask( hash_table, mask );
     return ;
}

/*===================================================================*/

void  PARTI_clear_stamp ( hash_table, stamp )
HASHSTR *hash_table;
int      stamp;
{
     clear_mask( hash_table,  PARTI_make_mask(stamp) );
}

/*===================================================================*/

int   find_inc_size(hash_table, combo_mask)
HASHSTR *hash_table;
int      combo_mask;
{
   int  list[ MAX_STAMPS ] ;
   int  i,siz=0;

   for(i=0; i<MAX_STAMPS; list[i]=0,i++);
   read_mask( combo_mask, list );
   for ( i= 0; i< MAX_STAMPS; i++ )  {
      if ( list[i] > 0 ) { 
      	siz  += hash_table->area_stamp[i];
      }
   }
   return siz; 
}
/*===============================================================*/
int   find_size(hash_table, combo_mask)
HASHSTR *hash_table;
int      combo_mask;
{
   int  list[ MAX_STAMPS ] ;
   int  i,siz=0;

   for(i=0; i<MAX_STAMPS; list[i]=0,i++);
   read_mask( combo_mask, list );
   for ( i= 0; i< MAX_STAMPS; i++ )  {
      if ( list[i] > 0 ) { 
      	siz  += hash_table->size_stamp[i];
      }
   }
   return siz; 
}

/*==================================================================*/
int   check_contiguity(hash_table, combo_mask)
HASHSTR *hash_table;
int      combo_mask;
{
   int  list[ MAX_STAMPS ] ;
   int  i,siz=0;
   int  st, end, cont;

   for(i=0; i<MAX_STAMPS; list[i]=0,i++);
   read_mask( combo_mask, list );

   for ( i=0, st=0, end=0, cont = 1; i< MAX_STAMPS; i++ ) {
       if (list[i] > 0 ) { 
         if ( !st ) st = 1;
         else if ( st && end ) {
	    cont = 0;
            break;
         } 
       }
       if (list[i] == 0 )    
           if ( st ) end = 1; 
   }
            
   return cont; 
}

/*======================================================================*/
int   estimate_hash_table_size( data_range )
int   data_range;
{
  int    j;
  double k ;

  if ( HARDWIRED_HASH_SIZE ) j =  HARDWIRED_HASH_SIZE ;
  else {
     k = (double)  PERCENT_DATA_RANGE / 100;
     j = (int)( (data_range/PARTI_numnodes )  * k );
  }

  if ( j < 64 ) j = 64;

  return j;
}
/*=======================================================================*/

